import inspect
import time
import traceback
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Tuple, Union

from junit_xml import TestCase

from ._junit_decorator import JunitDecorator
from ._junit_test_suite import JunitTestSuite


@dataclass
class CaseFailure:
    message: str
    output: str = ""
    type: str = ""

    def __getitem__(self, item):
        return self.__getattribute__(item)


@dataclass
class TestCaseData:
    case: Union[TestCase, None]
    func: Union[Callable, None]
    start_time: float
    parametrize: Union[None, List[Tuple[str, Any]]] = None

    @property
    def name(self):
        return self.func.__name__

    def set_fin_time(self):
        self.case.elapsed_sec = time.time() - self.start_time

    def get_case_key(self):
        if self.parametrize:
            return tuple(sorted(self.parametrize))
        return ()


class TestCaseCategories:
    FUNCTION = "function"
    FIXTURE = "fixture"


class JunitTestCase(JunitDecorator):
    """
    JunitTestCase is a decorator that represents single TestCase.
    When the decorated function finished its execution, it's registered to the relevant JunitTestSuite.
    TestCase will fail (TestCase failure) only when exception occurs during execution.
    """

    _func: Union[Callable, None]

    def __init__(self) -> None:
        super().__init__()

    @property
    def name(self):
        if self._func:
            return self.data.name

    def _on_wrapper_start(self, function):
        start_time = time.time()
        case = TestCase(name=function.__name__, classname=self._get_class_name(), category=TestCaseCategories.FUNCTION)
        self.data = TestCaseData(start_time=start_time, case=case, func=function)

    def _on_exception(self, e: BaseException):
        failure = CaseFailure(message=str(e), output=traceback.format_exc(), type=e.__class__.__name__)
        self.data.case.failures.append(failure)
        raise

    def _on_wrapper_end(self):
        self.data.set_fin_time()
        JunitTestSuite.register_case(self.data, self.get_suite_key())
        if self.data.parametrize:
            self.data.case.name += f'({", ".join([f"{p[0]}={p[1]}" for p in self.data.parametrize])})'

    def get_suite_key(self) -> Callable:
        """
        Get suite function as unique key.
        This function handles also on case that the test suite function decorated with pytest.mark.parametrize and
        collect its parameters and record them into the test case.
        :return: Wrapped Suite function instance
        """
        stack_locals = [frame_info.frame.f_locals for frame_info in inspect.stack()]
        suite_func = self._get_class_suite(stack_locals)
        if suite_func is None:
            suite_func = self._get_module_suite(stack_locals)
        return suite_func

    def _get_class_suite(self, stack_locals: List[Dict[str, Any]]):
        suite_func = None

        for f_locals in [
            stack_local for stack_local in stack_locals if "function" in stack_local and "self" in stack_local
        ]:
            suite = f_locals["self"]
            if isinstance(suite, JunitTestSuite):
                suite_func = f_locals["function"]
                stack_keys = list(inspect.signature(suite_func).parameters.keys())
                suite_arguments = [stack for stack in stack_locals if all(key in stack for key in stack_keys)][0]

                self.__set_parametrize(suite_func, suite_arguments)
                return suite_func

        return suite_func

    def _get_module_suite(self, stack_locals: List[Dict[str, Any]]):
        suite_func = None

        for f_locals in [
            stack_local
            for stack_local in stack_locals
            if "item" in stack_local and "nextitem" in stack_local and stack_local["item"].name == self.data.name
        ]:
            module = f_locals["nextitem"].parent
            suite_func = getattr(module.obj, f_locals["nextitem"].name).__wrapped__
            return suite_func

        return suite_func

    def __set_parametrize(self, suite_func, suite_arguments):
        if hasattr(suite_func, "pytestmark"):
            prams = list()
            marks = [m for m in suite_func.pytestmark if m.name == "parametrize"]

            for i in range(len(marks)):
                arg = marks[len(marks) - i - 1].args[0]
                prams.append((arg, suite_arguments[arg]))
            self.data.parametrize = prams



import os
from pathlib import Path
from typing import Any, Callable, ClassVar, Dict, List, Union

from junit_xml import TestCase, TestSuite, to_xml_report_string

from ._junit_decorator import JunitDecorator


class SuiteNotExistError(KeyError):
    """ Test Suite decorator name is not exists in suites poll """


class DuplicateSuiteError(KeyError):
    """ Test Suite decorator name is already exist in suites poll """


class JunitTestSuite(JunitDecorator):
    """
    JunitTestSuite is a decorator that waits for JunitTestCases decorators to register their testcases instances.
    After all test cases has finished their execution, the JunitTestSuite instance collect all registered TestCases
    and generate junit xml accordingly.
    JunitTestSuite represents single TestSuite.
    Generated report format: junit_<Class Name>_<Suite Function Name>_report.xml
    The usage of this function is regardless of pytest unittest packages.
    Default report path can be override if DEFAULT_REPORT_PATH_KEY environment variable is set
    """

    _junit_suites: ClassVar[Dict[Callable, Dict]] = dict()
    _report_dir: Path
    _cases = List[TestCase]
    _func: Union[Callable, None]
    _suite: Union[TestSuite, None]

    DEFAULT_REPORT_PATH_KEY = "JUNIT_REPORT_DIR"
    FAIL_ON_MISSING_SUITE_KEY = "FAIL_ON_MISSING_SUITE"

    XML_REPORT_FORMAT = "junit_{suite_name}_report{args}.xml"
    FAIL_ON_MISSING_SUITE = os.getenv(FAIL_ON_MISSING_SUITE_KEY, "False").lower() in ["true", "1", "yes", "y"]

    def __init__(self, report_dir: Path = None):
        """
        :param report_dir: Target directory, created if not exists
        """
        super().__init__()
        self._report_dir = self.get_report_dir(report_dir)
        self._cases = list()
        self._klass = None

    def _on_call(self):
        self._register()

    def _on_wrapper_end(self):
        self._collect(self._get_class_name())

    @classmethod
    def get_report_dir(cls, report_dir: Union[Path, None]) -> Path:
        if report_dir is None:
            return Path(os.getenv(cls.DEFAULT_REPORT_PATH_KEY, Path.cwd()))
        return report_dir

    @classmethod
    def collect_all(cls):
        """
        Collect all junit test suites reports from external source
        :return: None
        """
        for junit_suite in cls._junit_suites.values():
            suite = TestSuite(
                name=f"fixture_{junit_suite.name}",
                test_cases=cls._get_cases(junit_suite),
            )

            cls._export(junit_suite, suite)

    @classmethod
    def is_suite_exist(cls, suite_func: Callable):
        return suite_func in cls._junit_suites

    @classmethod
    def register_case(cls, test_data, suite_func: Callable) -> None:
        """
        Register test case to the relevant test suite
        :param test_data: TestCaseData instance
        :param suite_func: Wrapped function as cases key
        :return: None
        """
        if cls.is_suite_exist(suite_func):
            cls._add_case(cls._junit_suites[suite_func], test_data)
        else:
            if cls.FAIL_ON_MISSING_SUITE:
                raise SuiteNotExistError(f"Can't find suite named {suite_func} for {test_data} test case")

    def _execute_wrapped_function(self, *args, **kwargs) -> Any:
        return super()._execute_wrapped_function(*args, **kwargs)

    def _register(self):
        if self._func in JunitTestSuite._junit_suites:
            raise DuplicateSuiteError(f"Suite {self.name} already exist")
        JunitTestSuite._junit_suites[self._func] = self

    def _get_cases(self):
        return [data.case for data in self._cases]

    def _add_case(self, test_data):
        return self._cases.append(test_data)

    def _collect(self, class_name: str):
        """
        Collect all TestCases that
        :param class_name: Class name of which the decorated function contained in it
        :return: None
        """
        suite = TestSuite(
            name=f"{class_name}_{self.name}",
            test_cases=self._get_cases(),
        )
        self._export(suite)

    def _get_parametrize_values(self):
        values = ""
        parametrize = [p for p in self._cases if p.parametrize is not None]
        if parametrize:
            marks = {c.get_case_key() for c in self._cases if len(c.get_case_key()) > 0}
            if marks:
                values = "_".join([str(tup[1]) for tup in marks.pop()])
        return values

    def _export(self, suite: TestSuite) -> None:
        if len(suite.test_cases) == 0:
            return
        values = self._get_parametrize_values()
        path = self._report_dir.joinpath(self.XML_REPORT_FORMAT.format(suite_name=suite.name, args=values))
        xml_string = to_xml_report_string([suite])

        os.makedirs(self._report_dir, exist_ok=True)
        with open(path, "w") as f:
            f.write(xml_string)

        self._cases = list()
